import typing as t

from pydantic import BaseModel

from .base import settings


def register_settings(key: str) -> t.Callable[[t.Type[BaseModel]], t.Type[BaseModel]]:
    """
    Decorator to register a settings class with the settings module.

    Args:
        key: The key to register the settings under in the settings module.

    Usage:

    ```python
    @register_settings("MY_SETTINGS")
    class MySettings(BaseModel):
        foo: str
        bar: int
    ```

    ```python
    # settings.py
    MY_SETTINGS = {
        "foo": "bar",
        "bar": 1,
    }
    ```

    ```python
    from unfazed.conf import settings
    my_settings: MySettings = settings["MY_SETTINGS"]
    ```

    """

    def decorator(cls: t.Type[BaseModel]) -> t.Type[BaseModel]:
        settings.register_client_cls(key, cls)

        return cls

    return decorator



import inspect
import typing as t
from importlib import import_module

from unfazed.type import CanBeImported, HttpMethod

from .registry import _flatten_patterns
from .routing import Route, Static


def include(route_path: str) -> t.List[Route]:
    """
    Convert a route module path to a list of Route.

    Raises:
        ValueError: If the route module does not have a 'patterns' list.
        ValueError: If the route module does not have a list of Route.

    Usage:

    ```python

    from unfazed.route import include, path

    patterns = [
        path("/foo", routes=include("module.routes")),
    ]


    ```

    """
    route_module = import_module(route_path)
    app_label = route_path.rsplit(".", 1)[0]

    if not hasattr(route_module, "patterns"):
        raise ValueError(f"{route_path} must have a 'patterns' list")

    patterns = _flatten_patterns(route_module.patterns)
    for route in patterns:
        route.update_label(app_label)

    return patterns


@t.overload
def path(
    path: str,
    *,
    endpoint: t.Callable,
    methods: t.List[HttpMethod] | None = None,
    name: str | None = None,
    app_label: str | None = None,
    middlewares: t.List[CanBeImported] | None = None,
    include_in_schema: bool = True,
    tags: t.List[str] | None = None,
    summary: str | None = None,
    description: str | None = None,
    externalDocs: t.Dict | None = None,
    deprecated: bool = False,
    operation_id: str | None = None,
) -> Route: ...


@t.overload
def path(
    path: str,
    *,
    routes: t.List[Route] | None = None,
    methods: t.List[HttpMethod] | None = None,
    name: str | None = None,
    app_label: str | None = None,
    middlewares: t.List[CanBeImported] | None = None,
    include_in_schema: bool = True,
    tags: t.List[str] | None = None,
    summary: str | None = None,
    description: str | None = None,
    externalDocs: t.Dict | None = None,
    deprecated: bool = False,
    operation_id: str | None = None,
) -> t.List[Route]: ...


def path(
    path: str,
    *,
    endpoint: t.Callable | None = None,
    routes: t.List[Route] | None = None,
    methods: t.List[HttpMethod] | None = None,
    name: str | None = None,
    app_label: str | None = None,
    middlewares: t.List[CanBeImported] | None = None,
    include_in_schema: bool = True,
    tags: t.List[str] | None = None,
    summary: str | None = None,
    description: str | None = None,
    externalDocs: t.Dict | None = None,
    deprecated: bool | None = None,
    operation_id: str | None = None,
) -> Route | t.List[Route]:
    """

    Create a Route or a list of Route.

    Raises:
        ValueError: If exactly one of 'endpoint' or 'routes' is not provided.
        ValueError: If 'endpoint' is not a function.

    Usage:

    ```python
    from unfazed.route import path, include

    routes = [
        path("/foo", endpoint=home),
        path("/bar", routes=subpaths),
        path("/foobar", routes=include("module.routes.routes")),
    ]
    ```
    """
    if not (bool(endpoint) ^ bool(routes)):
        raise ValueError(
            f"error for {path}: Exactly one of 'endpoint' or 'routes' be provided"
        )

    if endpoint:
        if inspect.isfunction(endpoint):
            return Route(
                path,
                endpoint,
                methods=methods,
                name=name,
                middlewares=middlewares,
                app_label=app_label,
                include_in_schema=include_in_schema,
                tags=tags,
                summary=summary,
                description=description,
                externalDocs=externalDocs,
                deprecated=deprecated,
                operation_id=operation_id,
            )

        else:
            raise ValueError(f"endpoint {endpoint} must be a function")

    elif routes:
        ret = []
        for route in routes:
            if not isinstance(route, Route):
                raise ValueError(f"error for {path}: routes should be a list of Route")

            route.update_path(path + route.path)
            if middlewares:
                route.load_middlewares(middlewares)

            if not route.app_label:
                route.app_label = app_label

            ret.append(route)

        return ret

    # comply with mypy check
    # never reach here
    else:
        return []  # pragma: no cover


def static(
    path: str,
    directory: str,
    *,
    name: str | None = None,
    app_label: str | None = None,
    html: bool = False,
    include_in_schema: bool = True,
    tags: t.List[str] | None = None,
    summary: str | None = None,
    description: str | None = None,
    externalDocs: t.Dict | None = None,
    deprecated: bool = False,
    operation_id: str | None = None,
) -> Static:
    return Static(
        path=path,
        directory=directory,
        name=name,
        html=html,
        app_label=app_label,
        include_in_schema=include_in_schema,
        tags=tags,
        summary=summary,
        description=description,
        externalDocs=externalDocs,
        deprecated=deprecated,
        operation_id=operation_id,
    )



from .asgi import ASGIType
from .cache import CompressorBase, SerializerBase
from .middleware import MiddleWare
from .orm import DataBaseDriver, Model, QuerySet

__all__ = [
    "MiddleWare",
    "DataBaseDriver",
    "Model",
    "QuerySet",
    "ASGIType",
    "SerializerBase",
    "CompressorBase",
]



import typing as t

from openapi_pydantic.v3.v3_0 import ExternalDocumentation
from openapi_pydantic.v3.v3_0 import Info as InfoV30
from openapi_pydantic.v3.v3_0 import Server as ServerV30
from pydantic import BaseModel, Field


class SwaggerUI(BaseModel):
    css: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui.css"
    js: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@5/swagger-ui-bundle.js"
    favicon: str = (
        "https://www.openapis.org/wp-content/uploads/sites/3/2016/11/favicon.png"
    )


class Redoc(BaseModel):
    js: str = "https://cdn.redoc.ly/redoc/latest/bundles/redoc.standalone.js"


class OpenAPI(BaseModel):
    # openapi schema config
    openapi: t.Literal["3.0.0", "3.0.1", "3.0.2", "3.0.3", "3.0.4"] = "3.0.4"
    servers: t.List[ServerV30] = [ServerV30(url="/")]
    info: InfoV30
    externalDocs: ExternalDocumentation | None = None

    # openapi ui config
    json_route: str = "/openapi/openapi.json"
    swagger_ui: SwaggerUI = SwaggerUI()
    redoc: Redoc = Redoc()

    # open to network
    allow_public: bool = Field(
        default=True,
        description="allow public access, set to False when deploy to production",
    )



import logging
import typing as t
from contextlib import asynccontextmanager

from unfazed.type import ASGIApp

from .base import BaseLifeSpan

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover

logger = logging.getLogger("unfazed.setup")


class State(t.Dict[str, t.Any]):
    """A dictionary-like class to store application state."""

    pass


class LifeSpanHandler:
    """
    Handles the lifecycle events (startup and shutdown) for various components.

    This class manages the registration and execution of lifespan events for different
    parts of the application.
    """

    def __init__(self) -> None:
        """Initialize the lifespan handler with empty collections."""
        self.lifespan: t.Dict[str, BaseLifeSpan] = {}
        self.unfazed: "Unfazed" | None = None

    def register(self, name: str, lifespan: BaseLifeSpan) -> None:
        """
        Register a new lifespan component.

        Args:
            name: The name of the lifespan component
            lifespan: The lifespan component to register

        Raises:
            ValueError: If a lifespan with the same name is already registered
        """
        if name in self.lifespan:
            raise ValueError(f"Lifespan component '{name}' is already registered")
        self.lifespan[name] = lifespan

    def get(self, name: str) -> BaseLifeSpan | None:
        """
        Get a registered lifespan component by name.

        Args:
            name: The name of the lifespan component

        Returns:
            The lifespan component if found, None otherwise
        """
        return self.lifespan.get(name)

    async def on_startup(self) -> None:
        """
        Execute the startup event for all registered lifespan components.

        This method is called when the application starts.
        """
        for _, lifespan in self.lifespan.items():
            try:
                await lifespan.on_startup()
            except Exception as e:
                # Log the error but continue with other components
                logger.error(f"Error during startup: {e}")
                raise e

    async def on_shutdown(self) -> None:
        """
        Execute the shutdown event for all registered lifespan components.

        This method is called when the application shuts down.
        """
        for _, lifespan in self.lifespan.items():
            try:
                await lifespan.on_shutdown()
            except Exception as e:
                # Log the error but continue with other components
                logger.error(f"Error during shutdown: {e}")
                raise e

    @property
    def state(self) -> State:
        """
        Get the combined state from all lifespan components.

        Returns:
            A State object containing the combined state
        """
        ret = State()
        for _, lifespan in self.lifespan.items():
            ret.update(lifespan.state)
        return ret

    def clear(self) -> None:
        """Clear all registered lifespan components."""
        self.lifespan.clear()


# Global instance of the lifespan handler
handler = LifeSpanHandler()


@asynccontextmanager
async def lifespan(app: ASGIApp) -> t.AsyncIterator[State]:
    """
    ASGI lifespan context manager.

    This function is used as a lifespan context manager for ASGI applications.
    It handles the startup and shutdown events for all registered lifespan components.

    Args:
        app: The ASGI application

    Yields:
        The application state
    """
    await handler.on_startup()
    yield handler.state
    await handler.on_shutdown()



import http.cookies
import typing as t


def build_cookie(
    key: str,
    value: str,
    *,
    max_age: int | None = None,
    expires: str | None = None,
    path: str | None = "/",
    domain: str | None = None,
    secure: bool = False,
    httponly: bool = True,
    samesite: t.Literal["lax", "strict", "none"] | None = "lax",
) -> str:
    cookie: http.cookies.BaseCookie[str] = http.cookies.SimpleCookie()
    cookie[key] = value
    if max_age is not None:
        cookie[key]["max-age"] = max_age
    if expires is not None:
        cookie[key]["expires"] = expires
    if path is not None:
        cookie[key]["path"] = path
    if domain is not None:
        cookie[key]["domain"] = domain
    if secure:
        cookie[key]["secure"] = True
    if httponly:
        cookie[key]["httponly"] = True
    if samesite is not None:
        cookie[key]["samesite"] = samesite
    cookie_val = cookie.output(header="").strip()
    return cookie_val



import typing as t

from pydantic import BaseModel

from unfazed.schema import Condition
from unfazed.type import Doc


class Detail(BaseModel):
    name: str
    data: t.Dict[str, t.Any]


class Data(BaseModel):
    cond: t.List[Condition] = []
    name: str
    page: int
    size: int


class Action(BaseModel):
    name: str
    action: str
    data: t.Dict[str, t.Any]


class Save(BaseModel):
    name: str
    data: t.Annotated[
        t.Dict[str, t.Any],
        Doc(
            description="depends on the tortoise model, use unfazed.contrib.auth.models.User as example",
            examples=[
                {
                    "id": 1,
                    "username": "admin",
                    "email": "",
                    "created_at": 12345,
                    "updated_at": 12345,
                }
            ],
        ),
    ]
    inlines: t.Annotated[
        t.Dict[str, t.List[t.Dict[str, t.Any]]],
        Doc(
            description="relation model to `data`, use unfazed.contrib.auth.models.Group as example",
            examples={
                "groups": [{"name": "group1", "id": 1}, {"name": "group2", "id": 2}],
                "roles": [{"name": "role1", "id": 1}, {"name": "role2", "id": 2}],
            },
        ),
    ]


class Delete(BaseModel):
    name: str
    data: t.Dict[str, t.Any]



import typing as t
from abc import ABC, abstractmethod

from unfazed.contrib.auth.models import AbstractUser
from unfazed.contrib.auth.schema import LoginCtx, RegisterCtx
from unfazed.type import Doc, GenericReponse


class BaseAuthBackend(ABC):
    def __init__(self, options: t.Dict | None = None) -> None:
        self.options = options or {}

    @property
    @abstractmethod
    def alias(self) -> str: ...

    @abstractmethod
    async def login(
        self, ctx: LoginCtx
    ) -> t.Tuple[
        t.Annotated[
            t.Dict,
            Doc(description="session info, it will be set in the request.session"),
        ],
        t.Annotated[GenericReponse, Doc(description="response data")],
    ]: ...

    @abstractmethod
    async def session_info(
        self, user: AbstractUser, ctx: LoginCtx
    ) -> t.Dict[str, t.Any]: ...

    @abstractmethod
    async def register(self, ctx: RegisterCtx) -> None: ...

    @abstractmethod
    async def logout(self, session: t.Dict[str, t.Dict]) -> t.Any: ...



import inspect
import typing as t
import warnings
from functools import wraps

from starlette.concurrency import run_in_threadpool

from .handler import caches

P = t.ParamSpec("P")


def cached(
    using: str = "default",
    timeout: int = 60,
    include: t.List[str] | None = None,
) -> t.Callable:
    """
    Decorator for caching the results of async or sync functions.

    This decorator caches the return value of a function based on its arguments.
    The cache key is generated using the function's module, name, and specified parameters.

    Args:
        using (str): Name of the cache backend to use. Must be configured in settings.CACHES.
                    Defaults to "default".
        timeout (int): Time in seconds before the cache entry expires. Defaults to 60.
        include (List[str] | None): List of parameter names to include in the cache key.
                                  If None, all parameters are included. Defaults to None.

    Returns:
        Callable: A decorated async function that caches its results.

    Raises:
        KeyError: If the specified cache backend is not available.

    Examples:
        Basic usage:
        ```python
        @cached(timeout=60)
        async def get_user_info(user_id: int) -> dict:
            return {"user_id": user_id, "name": "Alice"}
        ```

        Using specific parameters for cache key:
        ```python
        @cached(timeout=60, include=["user_id"])
        async def get_user_info(user_id: int, name: str) -> dict:
            return {"user_id": user_id, "name": name}
        ```

        Using a different cache backend:
        ```python
        @cached(using="redis", timeout=60)
        async def get_user_info(user_id: int) -> dict:
            return {"user_id": user_id, "name": "Alice"}
        ```

        Force cache update:
        ```python
        # This will bypass the cache and update it with new data
        await get_user_info(user_id=1, force_update=True)
        ```

    Notes:
        - The decorator only works with keyword arguments. Positional arguments are ignored.
        - Cache keys are generated using the format: "module:function_name:param1_value:param2_value"
        - The cache can be bypassed using the `force_update=True` parameter.
    """

    def decorator(
        func: t.Callable[P, t.Awaitable[t.Any] | t.Any],
    ) -> t.Callable[P, t.Awaitable[t.Any] | t.Any]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
            if args:
                warnings.warn(
                    "The cached decorator will ignore positional arguments, use keyword arguments instead.",
                    UserWarning,
                    stacklevel=2,
                )

            prefix = f"{func.__module__}:{func.__name__}"
            if include:
                suffix = ":".join(
                    [f"{k}_{str(v)}" for k, v in kwargs.items() if k in include]
                )

            else:
                suffix = ""

            key = f"{prefix}:{suffix}" if suffix else prefix

            force_update = kwargs.pop("force_update", False)

            cache = caches[using]

            value = await cache.get(key)
            if value and not force_update:
                return value
            else:
                if inspect.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = await run_in_threadpool(func, *args, **kwargs)
                await cache.set(key, result, timeout)
                return result

        return wrapper

    return decorator



from unfazed.route import path

from .endpoints import docs, openapi_json, redoc

patterns = [
    path("/openapi/redoc", endpoint=redoc, include_in_schema=False),
    path("/openapi/docs", endpoint=docs, include_in_schema=False),
    path("/openapi/openapi.json", endpoint=openapi_json, include_in_schema=False),
]



import typing as t
from asyncio import Lock

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


def unfazed_locker(async_method: t.Callable) -> t.Callable:
    """
    Docorator to help Unfazed to lock the method until it's ready.

    """

    async def wrapper(self: "Unfazed", *args: t.Any, **kwargs: t.Any) -> t.Any:
        if self._ready:
            return
        async with Lock():
            if self._ready:
                return  # pragma: no cover

            if self._loading:
                raise RuntimeError("Unfazed is already loading")

            self._loading = True

            ret = await async_method(self, *args, **kwargs)

            self._ready = True
            self._loading = False

            return ret

    return wrapper



import typing as t

from pydantic import BaseModel

from unfazed.route import params


class LoginCtx(BaseModel):
    account: str = ""
    password: str = ""

    # for most oauth login
    token: str = ""
    platform: str = "default"

    extra: t.Dict[str, t.Any] = {}


class LoginResponse(BaseModel):
    extra: t.Dict[str, t.Any] | None = None


class LoginSucceed(BaseModel):
    msg: str = ""
    code: int = 0
    data: LoginResponse | None = None


class LogoutSucceed(BaseModel):
    msg: str = ""
    code: int = 0


LOGIN_RESPONSE = [params.ResponseSpec(model=LoginSucceed)]
LOUOUT_RESPONSE = [params.ResponseSpec(model=LogoutSucceed)]


class RegisterCtx(BaseModel):
    account: str = ""
    password: str = ""

    platform: str = "default"
    extra: t.Dict[str, t.Any] = {}



import typing as t
import uuid
from pathlib import Path

from click import Option
from jinja2 import Template

import unfazed
from unfazed.command import BaseCommand


class Command(BaseCommand):
    help_text = """

    Create a new project from cli

    Usage:
    >>> unfazed-cli startproject -n myproject

    """

    @t.override
    def add_arguments(self) -> t.List[Option]:
        return [
            Option(
                ["--project_name", "-n"],
                help="name of the project",
                type=str,
            ),
            Option(
                ["--location", "-l"],
                help="location of the project",
                type=str,
                default=str(Path.cwd()),
                show_default=True,
            ),
        ]

    async def handle(self, **options: t.Any) -> None:
        project_name = options["project_name"]
        location = options["location"]
        template_path = Path(unfazed.__path__[0], "template/project")

        location_path = Path(location)
        project_path = location_path / project_name

        if project_path.exists():
            raise FileExistsError(f"Project {project_path} already exists")

        context = {
            "project_name": project_name,
            "secret": uuid.uuid4().hex + uuid.uuid4().hex,
        }

        for root, _, files in template_path.walk():
            for f in files:
                new_root = location_path / root.relative_to(template_path)
                self.handle_file(root, new_root, f, context)

    def handle_file(
        self,
        root: Path,
        new_root: Path,
        file_path: str,
        context: dict,
    ) -> None:
        new_root_rendered = Path(Template(str(new_root)).render(**context))
        if not new_root_rendered.exists():
            new_root_rendered.mkdir(parents=True, exist_ok=True)

        cur_file = root.joinpath(file_path)
        if file_path.endswith(".py.tpl"):
            file_path = file_path[:-4]
        new_file = new_root_rendered.joinpath(file_path)
        with open(cur_file, "r") as cur_fh, open(new_file, "w") as new_fh:
            rendered_text = Template(cur_fh.read(), autoescape=True).render(**context)
            new_fh.write(rendered_text)



from .collector import admin_collector
from .decorators import action, register
from .models import (
    BaseAdmin,
    BaseModelAdmin,
    CacheAdmin,
    ModelAdmin,
    ModelInlineAdmin,
    ToolAdmin,
    site,
)
from .utils import parse_cond

__all__ = [
    "admin_collector",
    "BaseAdmin",
    "ModelAdmin",
    "parse_cond",
    "site",
    "register",
    "action",
    "ModelInlineAdmin",
    "CacheAdmin",
    "ToolAdmin",
    "BaseModelAdmin",
]



import typing as t

from pydantic import BaseModel, Field


class Cors(BaseModel):
    """
    refer https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
    """

    allow_origins: t.List[str] = Field(
        default=["*"], description="List of allowed origins", alias="ALLOW_ORIGINS"
    )
    allow_methods: t.List[str] = Field(
        default=["*"], description="List of allowed methods", alias="ALLOW_METHODS"
    )
    allow_headers: t.List[str] = Field(
        default=[], description="List of allowed headers", alias="ALLOW_HEADERS"
    )
    allow_credentials: bool = Field(
        default=False, description="Allow credentials", alias="ALLOW_CREDENTIALS"
    )

    allow_origin_regex: str | None = Field(
        default=None, description="Allow origin regex", alias="ALLOW_ORIGIN_REGEX"
    )
    expose_headers: t.List[str] = Field(
        default=[], description="List of exposed headers", alias="EXPOSE_HEADERS"
    )
    max_age: int = Field(default=600, description="Max age", alias="MAX_AGE")


class TrustedHost(BaseModel):
    allowed_hosts: t.List[str] = Field(
        default=["*"], description="List of allowed hosts", alias="ALLOWED_HOSTS"
    )

    www_redirect: bool = Field(
        default=True, description="Redirect www to non-www", alias="WWW_REDIRECT"
    )


class GZip(BaseModel):
    minimum_size: int = Field(
        default=500, description="Minimum size for compression", alias="MINIMUM_SIZE"
    )

    compress_level: int = Field(
        default=9, description="Compression level", alias="COMPRESS_LEVEL"
    )



from unfazed.contrib.admin.registry import ModelAdmin, register

from .serializers import (
    GroupSerializer,
    PermissionSerializer,
    RoleSerializer,
    UserSerializer,
)


@register(UserSerializer)
class UserAdmin(ModelAdmin):
    inlines = ["InlineGroupAdmin", "InlineRoleAdmin"]


@register(UserSerializer)
class InlineUserAdmin(ModelAdmin):
    pass


@register(GroupSerializer)
class GroupAdmin(ModelAdmin):
    inlines = ["InlineRoleAdmin", "InlineUserAdmin"]


@register(GroupSerializer)
class InlineGroupAdmin(ModelAdmin):
    pass


@register(RoleSerializer)
class RoleAdmin(ModelAdmin):
    inlines = ["PermissionAdmin"]


@register(RoleSerializer)
class InlineRoleAdmin(ModelAdmin):
    pass


@register(PermissionSerializer)
class PermissionAdmin(ModelAdmin):
    pass



import typing as t
from pathlib import Path

from aerich import Command as AerichCommand
from aerich.enums import Color
from click import Option, secho

from unfazed.command import BaseCommand


class Command(BaseCommand):
    help_text = "aerich: Init config file and generate root migrate location."

    def add_arguments(self) -> t.List[Option]:
        return [
            Option(
                ["--safe", "-s"],
                help="Safe mode",
                type=str,
                default=True,
                show_default=True,
            ),
            Option(
                ["--location", "-l"],
                help="Migrate store location.",
                type=str,
                default="./migrations",
                show_default=True,
            ),
        ]

    async def handle(self, **options: t.Any) -> None:
        location = options["location"]
        safe = options["safe"]

        app = "models"

        assert (
            self.unfazed.settings.DATABASE is not None
        ), "No database found in settings"

        db_conf = self.unfazed.settings.DATABASE.model_dump(exclude_none=True)
        aerich_cmd = AerichCommand(db_conf, location=location)

        dirname = Path(location, app)
        try:
            await aerich_cmd.init_db(safe)
            secho(f"Success create app migrate location {dirname}", fg=Color.green)
            secho(f'Success generate schema for app "{app}"', fg=Color.green)
        except FileExistsError:
            return secho(
                f"Inited {app} already, or delete {dirname} and try again.",
                fg=Color.yellow,
            )



from .base import BaseUnfazedException


class MethodNotAllowed(BaseUnfazedException):
    def __init__(self, message: str = "Method not allowed", code: int = 405):
        self.message = message
        self.code = code



import os
import typing as t
from importlib import import_module


def import_string(name: str) -> t.Type:
    """
    Import a class or object from a module using its fully qualified name.

    Args:
        name (str): The fully qualified name of the class or object to import.
                   Format should be 'module.submodule.ClassName'.

    Returns:
        t.Type: The imported class or object.

    Raises:
        ImportError: If the module path is invalid or the module cannot be imported.
        ValueError: If the name does not contain a dot (.) separator.

    Example:
        >>> import_string("unfazed.http.HttpResponse")
        <class 'unfazed.http.HttpResponse'>
    """

    try:
        module_name, class_name = name.rsplit(".", 1)
    except ValueError:
        raise ImportError(f"{name} doesn't look like a module path")

    module = import_module(module_name)
    return getattr(module, class_name)


def import_setting(env: str) -> t.Dict[str, t.Any]:
    """
    Import the settings module specified by the environment variable.

    Args:
        env (str): The name of the environment variable containing the settings module path.

    Returns:
        t.Dict[str, t.Any]: A dictionary containing all settings from the imported module.

    Raises:
        ValueError: If the environment variable is not set.
        ImportError: If the settings module cannot be imported.

    Example:
        >>> import_setting("UNFAZED_SETTINGS_MODULE")
        {'DEBUG': True, 'SECRET_KEY': '...', ...}
    """
    settings_module = os.environ.get(env)
    if not settings_module:
        raise ValueError(f"environment variable {env} is not set")
    try:
        settingskv = import_module(settings_module).__dict__
    except ImportError:
        raise ImportError(f"Could not import settings module {settings_module}")

    return settingskv



import zlib

from unfazed.protocol import CompressorBase


class ZlibCompressor(CompressorBase):
    def compress(self, value: bytes) -> bytes:
        return zlib.compress(value)

    def decompress(self, value: bytes) -> bytes:
        return zlib.decompress(value)



from .request import Action, Data, Delete, Detail, Save
from .response import (
    DATA_RESP,
    DELETE_RESP,
    DESC_RESP,
    DETAIL_RESP,
    SAVE_RESP,
    DataResp,
    DeleteResp,
    DescResp,
    DetailResp,
    SaveResp,
)

__all__ = [
    "DESC_RESP",
    "DETAIL_RESP",
    "DescResp",
    "DetailResp",
    "Detail",
    "Data",
    "DATA_RESP",
    "DataResp",
    "Action",
    "Save",
    "SAVE_RESP",
    "SaveResp",
    "Delete",
    "DELETE_RESP",
    "DeleteResp",
]



import typing as t
import warnings

import httpx
from asgiref.testing import ApplicationCommunicator
from httpx import Request, Response
from httpx._transports.asgi import ASGIResponseStream, create_event

from unfazed.type import ASGIApp, Receive, Scope, Send

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


# hope httpx.ASGITransport support state one day
# refer: https://github.com/encode/httpx/discussions/3464
# no cover: start
class ASGITransport(httpx.AsyncBaseTransport):
    """
    Add state support to httpx.ASGITransport
    """

    def __init__(
        self,
        app: ASGIApp,
        state: t.Dict[str, t.Any],
        raise_app_exceptions: bool = True,
        root_path: str = "",
        client: tuple[str, int] = ("127.0.0.1", 123),
    ) -> None:
        self.app = app
        self.raise_app_exceptions = raise_app_exceptions
        self.root_path = root_path
        self.client = client
        self.state = state

    async def handle_async_request(
        self,
        request: Request,
    ) -> Response:
        assert isinstance(request.stream, httpx.AsyncByteStream)

        # ASGI scope.
        scope = {
            "type": "http",
            "asgi": {"version": "3.0"},
            "http_version": "1.1",
            "method": request.method,
            "headers": [(k.lower(), v) for (k, v) in request.headers.raw],
            "scheme": request.url.scheme,
            "path": request.url.path,
            "raw_path": request.url.raw_path.split(b"?")[0],
            "query_string": request.url.query,
            "server": (request.url.host, request.url.port),
            "client": self.client,
            "root_path": self.root_path,
            "state": self.state.copy(),
        }

        # Request.
        request_body_chunks = request.stream.__aiter__()
        request_complete = False

        # Response.
        status_code = None
        response_headers = None
        body_parts = []
        response_started = False
        response_complete = create_event()

        # ASGI callables.

        async def receive() -> dict[str, t.Any]:
            nonlocal request_complete

            if request_complete:
                await response_complete.wait()
                return {"type": "http.disconnect"}

            try:
                body = await request_body_chunks.__anext__()
            except StopAsyncIteration:
                request_complete = True
                return {"type": "http.request", "body": b"", "more_body": False}
            return {"type": "http.request", "body": body, "more_body": True}

        async def send(message: dict[str, t.Any]) -> None:
            nonlocal status_code, response_headers, response_started

            if message["type"] == "http.response.start":
                assert not response_started

                status_code = message["status"]
                response_headers = message.get("headers", [])
                response_started = True

            elif message["type"] == "http.response.body":
                assert not response_complete.is_set()
                body = message.get("body", b"")
                more_body = message.get("more_body", False)

                if body and request.method != "HEAD":
                    body_parts.append(body)

                if not more_body:
                    response_complete.set()

        try:
            await self.app(scope, t.cast(Receive, receive), t.cast(Send, send))
        except Exception:  # noqa: PIE-786
            if self.raise_app_exceptions:
                raise

            response_complete.set()
            if status_code is None:
                status_code = 500
            if response_headers is None:
                response_headers = {}

        assert response_complete.is_set()
        assert status_code is not None
        assert response_headers is not None

        stream = ASGIResponseStream(body_parts)

        return Response(status_code, headers=response_headers, stream=stream)


# no cover: stop


class Requestfactory(httpx.AsyncClient):
    """
    A test client factory for Unfazed applications that provides an easy way to test your ASGI application.

    This class extends httpx.AsyncClient and provides a convenient interface for making HTTP requests
    to your Unfazed application during testing. It handles the ASGI transport and lifespan events
    automatically.

    Features:
        - Automatic lifespan management (startup/shutdown)
        - State sharing between requests
        - Simple HTTP request methods (get, post, put, etc.)
        - ASGI transport integration

    Usage:
        ```python
        from unfazed import Unfazed
        from unfazed.test.request import Requestfactory

        async def test_your_app() -> None:
            # Initialize your Unfazed application
            unfazed = Unfazed()
            await unfazed.setup()

            # Create a test client and make requests
            async with Requestfactory(unfazed) as request:
                # Make HTTP requests using httpx.AsyncClient methods
                response = await request.get("/")
                assert response.status_code == 200
                assert response.json() == {"message": "Hello, World!"}

                # Make POST request with data
                response = await request.post("/api/data", json={"key": "value"})
                assert response.status_code == 201
        ```

    Args:
        app: The Unfazed application instance to test
        lifespan_on: Whether to handle lifespan events (startup/shutdown). Defaults to True
        base_url: The base URL for all requests. Defaults to "http://testserver"
    """

    def __init__(
        self,
        app: "Unfazed",
        lifespan_on: bool = True,
        base_url: str = "http://testserver",
    ) -> None:
        self.app = app

        self.app_state: t.Dict[str, t.Any] = {}
        transport = ASGITransport(app, self.app_state)
        scope: Scope = {
            "type": "lifespan",
            "asgi": {"version": "3.0", "spec_version": "2.1"},
            "state": self.app_state,
        }
        self.communicator = ApplicationCommunicator(app, scope)  # type: ignore
        self.lifespan_on = lifespan_on
        super().__init__(base_url=base_url, transport=transport)

    async def __aenter__(self) -> t.Self:
        if self.lifespan_on:
            await self.lifespan_startup()

        return self

    async def __aexit__(self, *args: t.Any) -> None:
        if self.lifespan_on:
            await self.lifespan_shutdown()

    async def lifespan_startup(self) -> None:
        await self.communicator.send_input(
            {"type": "lifespan.startup", "state": self.app_state}
        )  # type: ignore
        message = await self.communicator.receive_output()  # type: ignore
        if message["type"] != "lifespan.startup.complete":
            raise RuntimeError("Startup failed")

    async def lifespan_shutdown(self) -> None:
        await self.communicator.send_input({"type": "lifespan.shutdown"})  # type: ignore
        message = await self.communicator.receive_output()  # type: ignore
        if message["type"] != "lifespan.shutdown.complete":
            warnings.warn("Shutdown failed", RuntimeWarning, stacklevel=2)



import typing as t

import orjson as json
from starlette.requests import Request

if t.TYPE_CHECKING:
    from unfazed.contrib.auth.models import AbstractUser  # pragma: no cover
    from unfazed.contrib.session.backends.base import SessionBase  # pragma: no cover
    from unfazed.core import Unfazed  # pragma: no cover


class HttpRequest(Request):
    """
    Enhanced HTTP request class for the Unfazed framework.

    This class extends Starlette's Request class to provide additional functionality
    specific to the Unfazed framework, including JSON parsing, session management,
    user authentication, and application access.

    Attributes:
        _json (t.Dict): Cached JSON body of the request.

    Properties:
        scheme (str): The URL scheme (http/https).
        path (str): The URL path.
        session (SessionBase): The session object if SessionMiddleware is installed.
        user (AbstractUser): The authenticated user if AuthenticationMiddleware is installed.
        unfazed (Unfazed): The Unfazed application instance.

    Example:
        ```python
        from unfazed.http import HttpRequest, HttpResponse

        async def my_view(request: HttpRequest):
            # Parse JSON request body
            body = await request.json()

            # Access request metadata
            path = request.path
            scheme = request.scheme

            # Access session (requires SessionMiddleware)
            session = request.session

            # Access authenticated user (requires AuthenticationMiddleware)
            user = request.user

            # Access the Unfazed application
            unfazed = request.unfazed

            return HttpResponse(content="Hello, world")
        ```
    """

    @t.override
    async def json(self) -> t.Dict:
        """
        Parse the request body as JSON using orjson.

        This method caches the parsed JSON to avoid parsing the body multiple times.

        Returns:
            t.Dict: The parsed JSON body.

        Raises:
            json.JSONDecodeError: If the request body is not valid JSON.
        """
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json

    @property
    def scheme(self) -> str:
        """
        Get the URL scheme (http/https).

        Returns:
            str: The URL scheme.
        """
        return self.url.scheme

    @property
    def path(self) -> str:
        """
        Get the URL path.

        Returns:
            str: The URL path.
        """
        return self.url.path

    @property
    @t.override
    def session(self) -> "SessionBase":  # type: ignore
        """
        Get the session object.

        Returns:
            SessionBase: The session object.

        Raises:
            ValueError: If SessionMiddleware is not installed.
        """
        if "session" not in self.scope:
            raise ValueError(
                "SessionMiddleware must be installed to access request.session"
            )

        return self.scope["session"]

    @property
    @t.override
    def user(self) -> "AbstractUser":  # type: ignore
        """
        Get the authenticated user.

        Returns:
            AbstractUser: The authenticated user or None if not authenticated.

        Raises:
            ValueError: If AuthenticationMiddleware is not installed.
        """
        if "user" not in self.scope:
            raise ValueError(
                "AuthenticationMiddleware must be installed to access request.user"
            )
        return self.scope.get("user", None)

    @property
    def unfazed(self) -> "Unfazed":
        """
        Get the Unfazed application instance.

        Returns:
            Unfazed: The Unfazed application instance.
        """
        return self.scope["app"]



import copy
import typing as t
from logging.config import dictConfig

if t.TYPE_CHECKING:
    from unfazed.core import Unfazed  # pragma: no cover


DEFAULT_LOGGING_CONFIG = {
    "formatters": {
        "_simple": {
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },
    "handlers": {
        "_console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "_simple",
        }
    },
    "loggers": {
        "unfazed": {
            "level": "DEBUG",
            "handlers": ["_console"],
        },
        "unvicorn": {
            "level": "INFO",
            "handlers": ["_console"],
        },
        "tortoise": {
            "level": "INFO",
            "handlers": ["_console"],
        },
    },
    "filters": {},
    # "root": {"level": "DEBUG", "handlers": ["_console"]},
    "version": 1,
}


class LogCenter:
    """Central logging configuration manager for Unfazed applications.

    This class handles the configuration and setup of logging for Unfazed applications.
    It provides a way to merge custom logging configurations with default settings
    and ensures proper logging setup.

    Attributes:
        unfazed: The Unfazed application instance
        raw_dictconfig: The raw logging configuration dictionary
        config: The merged logging configuration
    """

    def __init__(self, unfazed: "Unfazed", dictconfig: t.Dict[str, t.Any]) -> None:
        """Initialize the LogCenter with an Unfazed instance and logging configuration.

        Args:
            unfazed: The Unfazed application instance
            dictconfig: The logging configuration dictionary

        Raises:
            UnfazedSetupError: If the configuration is invalid
        """
        self.unfazed = unfazed
        self.raw_dictconfig = dictconfig
        self.config = self.merge_default(dictconfig)

    def setup(self) -> None:
        dictConfig(self.config)

    def merge_default(self, dictconfig: t.Dict) -> t.Dict:
        ret = copy.deepcopy(DEFAULT_LOGGING_CONFIG)

        if not dictconfig:
            return ret

        for key in dictconfig:
            if key not in ret:
                ret[key] = dictconfig[key]

            elif key in ["handlers", "formatters", "filters", "loggers"]:
                # ret[key].update(dictconfig[key])
                target = t.cast(t.Dict, ret[key])
                target.update(dictconfig[key])

            else:
                continue

        return ret



import os
import typing as t

from .asgi import ASGIApp, Message, Receive, Scope, Send
from .base import Doc
from .cache import CacheBackend, CacheOptions
from .http import (
    ContentStream,
    GenericReponse,
    HttpMethod,
)

CanBeImported = t.Annotated[
    str,
    Doc(
        description="can be imported by unfazed.utils.import_string",
        example="unfazed.core.Unfazed",
    ),
]
Domain = t.Annotated[
    str,
    Doc(
        description="host:port",
        example="127.0.0.1:9527",
    ),
]

PathLike = t.Union[str, "os.PathLike[str]"]


__all__ = [
    "CanBeImported",
    "CacheOptions",
    "Domain",
    "HttpMethod",
    "GenericReponse",
    "ContentStream",
    "CacheBackend",
    "PathLike",
    "Scope",
    "ASGIApp",
    "Message",
    "Receive",
    "Send",
]



import typing as t

from pydantic import BaseModel, Field

from unfazed.type import CanBeImported


class Cache(BaseModel):
    BACKEND: CanBeImported
    LOCATION: t.List[str] | str | None = None
    OPTIONS: t.Dict[str, t.Any] | None = None


class LocOptions(BaseModel):
    TIMEOUT: int | None = None
    PREFIX: str | None = None
    VERSION: int | None = None
    MAX_ENTRIES: int = 300


class RedisOptions(BaseModel):
    retry: bool = True

    socket_timeout: int | None = None
    socket_connect_timeout: int | None = None
    socket_keepalive: bool | None = None
    socket_keepalive_options: t.Mapping[int, t.Union[int, bytes]] | None = None

    # set decode responses to True
    decode_responses: bool = False
    retry_on_timeout: bool = False
    retry_on_error: t.List | None = None
    max_connections: int = 10
    single_connection_client: bool = False
    health_check_interval: int = 30

    ssl: bool = False
    ssl_keyfile: str | None = None
    ssl_certfile: str | None = None
    ssl_cert_reqs: str = "required"
    ssl_ca_certs: str | None = None
    ssl_ca_data: str | None = None
    ssl_check_hostname: bool = False
    ssl_min_version: t.Any = None
    ssl_ciphers: str | None = None

    serializer: CanBeImported | None = Field(
        "unfazed.cache.serializers.pickle.PickleSerializer",
        alias="SERIALIZER",
        description="serialize data before save",
    )

    compressor: CanBeImported | None = Field(
        "unfazed.cache.compressors.zlib.ZlibCompressor",
        alias="COMPRESSOR",
        description="compress data before save",
    )

    prefix: str | None = Field(
        None,
        alias="PREFIX",
        description="its strongly recommended to set prefix",
    )



from starlette.convertors import Convertor, register_url_convertor

from .base import include, path, static
from .registry import parse_urlconf
from .routing import Route

__all__ = [
    "Route",
    "path",
    "include",
    "parse_urlconf",
    "Convertor",
    "register_url_convertor",
    "static",
]



import typing as t

from unfazed.contrib.auth.decorators import login_required
from unfazed.contrib.common.utils import response_success
from unfazed.http import HttpRequest, HttpResponse, JsonResponse
from unfazed.route import params as p
from unfazed.type import Doc
from unfazed.utils import generic_response

from . import schema as s
from .decorators import record
from .models import LogEntry
from .services import AdminModelService


@login_required
async def list_route(request: HttpRequest) -> JsonResponse:
    ret = await AdminModelService.list_route(request)

    await LogEntry.create(
        created_at=1, account="1", path="/", ip="1", request="1", response="1"
    )
    return response_success(data=ret)


@login_required
def settings(request: HttpRequest) -> JsonResponse:
    ret = AdminModelService.site_settings()
    return response_success(data=ret)


@login_required
async def model_desc(
    request: HttpRequest, name: t.Annotated[str, p.Json()]
) -> t.Annotated[JsonResponse, *s.DESC_RESP]:
    ret = await AdminModelService.model_desc(name, request=request)
    return response_success(data=ret)


@login_required
async def model_detail(
    request: HttpRequest, ctx: t.Annotated[s.Detail, p.Json()]
) -> t.Annotated[JsonResponse, *s.DETAIL_RESP]:
    ret = await AdminModelService.model_detail(ctx.name, ctx.data, request=request)
    return response_success(data=ret)


@login_required
async def model_data(
    request: HttpRequest, ctx: t.Annotated[s.Data, p.Json()]
) -> t.Annotated[JsonResponse, *s.DATA_RESP]:
    ret = await AdminModelService.model_data(
        ctx.name, ctx.cond, ctx.page, ctx.size, request=request
    )
    return response_success(data=ret)


@login_required
@record
async def model_action(
    request: HttpRequest, ctx: t.Annotated[s.Action, p.Json()]
) -> t.Annotated[
    HttpResponse, Doc(description="depends on the action implemetion, text/json/stream")
]:
    ret = await AdminModelService.model_action(ctx.name, ctx.action, ctx.data, request)

    return generic_response(ret)


@login_required
@record
async def model_save(
    request: HttpRequest, ctx: t.Annotated[s.Save, p.Json()]
) -> t.Annotated[JsonResponse, *s.SAVE_RESP]:
    ret = await AdminModelService.model_save(
        ctx.name, ctx.data, ctx.inlines, request=request
    )

    return response_success(data=ret)


@login_required
@record
async def model_delete(
    request: HttpRequest, ctx: t.Annotated[s.Delete, p.Json()]
) -> t.Annotated[JsonResponse, *s.DELETE_RESP]:
    await AdminModelService.model_delete(ctx.name, ctx.data, request=request)
    return response_success()



import typing as t

from unfazed.conf import settings
from unfazed.contrib.auth.models import AbstractUser
from unfazed.contrib.auth.settings import UnfazedContribAuthSettings
from unfazed.type import ASGIApp, Receive, Scope, Send

if t.TYPE_CHECKING:
    from unfazed.contrib.session.backends.base import SessionBase  # pragma: no cover


class AuthenticationMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

        self.setting: UnfazedContribAuthSettings = settings[
            "UNFAZED_CONTRIB_AUTH_SETTINGS"
        ]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        if "user" in scope:
            await self.app(scope, receive, send)
            return

        session: "SessionBase" = t.cast("SessionBase", scope.get("session"))
        if session is None or (self.setting.SESSION_KEY not in session):
            user = None

        else:
            session_dict = session[self.setting.SESSION_KEY]
            UserCls = AbstractUser.UserCls()
            user = await UserCls.from_session(session_dict)

        scope["user"] = user

        await self.app(scope, receive, send)
